#include "echoes.h"

static char *_echoes_parse_nick_get(char *dest, char *src);
static char *_echoes_strstr(const char *haystack, const char *needle);
static int   _echoes_nick_compare(const void *d1, const void *d2);
static int   _echoes_strcase_compare(const void *d1, const void *d2);
static void  _echoes_nick_switch_cb(void *data, void *udata);
static void  _echoes_nick_quit_cb(void *data, void *udata);
static void  _echoes_strlist_concat_cb(void *data, void *fdata);

int
main(int argc, char **argv)
{
  Echoes *echoes;

  echoes = calloc(1, sizeof(Echoes));
  if (!etk_init(&argc, &argv))
    {
      fprintf(stderr, "error during etk initialisation\n");
      return EXIT_FAILURE;
    }
  enhance_init();
  ecore_con_init();

  /* Load the glade file */
  echoes->en = enhance_new();
  if (!access(PACKAGE_SOURCE_DATA_DIR "/glade/echoes.glade", R_OK))
    enhance_file_load(echoes->en, "window1", PACKAGE_SOURCE_DATA_DIR "/glade/echoes.glade");
  else
    enhance_file_load(echoes->en, "window1", PACKAGE_DATA_DIR "/glade/echoes.glade");

  /* Get the Widgets */
  echoes->server_list = enhance_var_get(echoes->en, "server_list");
  echoes->nick_list   = enhance_var_get(echoes->en, "nick_list");
  echoes->notebook    = enhance_var_get(echoes->en, "notebook");
  echoes->logo        = enhance_var_get(echoes->en, "logo");
  echoes->nick_button = enhance_var_get(echoes->en, "nick_button");
  echoes->msg_entry   = enhance_var_get(echoes->en, "msg_entry");

  /* Set the logo */
  if (!access(PACKAGE_SOURCE_DATA_DIR "/pixmaps/OOornithorynque.png", R_OK))
    etk_image_set_from_file(ETK_IMAGE(echoes->logo), PACKAGE_SOURCE_DATA_DIR "/pixmaps/OOornithorynque.png", NULL);
  else
    etk_image_set_from_file(ETK_IMAGE(echoes->logo), PACKAGE_DATA_DIR "/pixmaps/OOornithorynque.png", NULL);

  /* Set the trees */
  etk_tree_mode_set(ETK_TREE(echoes->server_list), ETK_TREE_MODE_TREE);
  Etk_Tree_Col *col;
  col = etk_tree_col_new(ETK_TREE(echoes->server_list), "Servers", 60, 0);
  etk_tree_col_model_add(col, etk_tree_model_image_new());
  etk_tree_col_model_add(col, etk_tree_model_text_new());
  etk_tree_build(ETK_TREE(echoes->server_list));
  col = etk_tree_col_new(ETK_TREE(echoes->nick_list), "Nicks", 60, 0);
  etk_tree_col_model_add(col, etk_tree_model_text_new());
  etk_tree_build(ETK_TREE(echoes->nick_list));

  /* Set the callbacks */
  etk_signal_connect("key-down", ETK_OBJECT(echoes->msg_entry), 
                     ETK_CALLBACK(on_msg_entry_key_down_event), echoes);
  etk_signal_connect("row-selected", ETK_OBJECT(echoes->server_list),
                     ETK_CALLBACK(on_server_list_row_select_event), echoes);
  etk_signal_connect("row-selected", ETK_OBJECT(echoes->nick_list),
                     ETK_CALLBACK(on_nick_list_row_select_event), echoes);
  etk_signal_connect("clicked", ETK_OBJECT(echoes->nick_button), ETK_CALLBACK(on_nick_button_clicked), echoes);

  //XXX Oui j'ai la flemme de coder une config
  echoes_defaults_set(echoes);

  etk_main();
  enhance_free(echoes->en);
  enhance_shutdown();
  etk_shutdown();

  return EXIT_SUCCESS;
}

void
echoes_defaults_set(Echoes *echoes)
{
  echoes->nick = strdup(IRC_NICK);
  echoes->chan_hash = ecore_hash_new(ecore_str_hash, _echoes_strcase_compare);
  ecore_hash_set_free_key(echoes->chan_hash, free);
  ecore_hash_set_free_value(echoes->chan_hash, echoes_server_free);
  echoes->serv_hash = ecore_hash_new(ecore_direct_hash, ecore_direct_compare);
  ecore_hash_set_free_key(echoes->serv_hash, echoes_server_disconnect);
  ecore_hash_set_free_value(echoes->serv_hash, echoes_server_free);
}

Ecore_List *
echoes_split_string(const char *msg, const char *delim)
{
  Ecore_List *list = ecore_list_new();
  char *token = NULL;
  char *str, *p;

  ecore_list_set_free_cb(list, free);
  p = str = strdup(msg);
  while ((token = strsep(&str, delim)))
    {
      ecore_list_append(list, strdup(token));
    }
  ecore_list_goto_first(list);
  free(p);

  return list;
}

#define ECHOES_IS_NOT_SPACE(x) (x && *x != '\0' && *x != ' ' && *x != '\r' && *x != '\n')
Ecore_List *
echoes_parse_message(const char *msg)
{
  const char *str;
  const char *p;
  char *prefix;
  char *command;
  char *params;
  char *trailing;
  Ecore_List *parsed_msg;
  parsed_msg = ecore_list_new();
  ecore_list_set_free_cb(parsed_msg, free);

  str = msg;
  /* Check for a prefix */
  if (*str == ':')
    {
      int prefix_len;

      for (p = str; ECHOES_IS_NOT_SPACE(p); p++);
      prefix_len = p - str;
      prefix = calloc(prefix_len+1, sizeof(char));
      strncpy(prefix, str, prefix_len);

      printf("prefix is %s\n", prefix);
      ecore_list_append(parsed_msg, prefix);
      for (str = p; str && *str == ' '; str++);
    }

  /* Get the command */
  {
    int command_len;

    for (p = str; ECHOES_IS_NOT_SPACE(p) && *p ; p++);
    if ((command_len = p - str))
      {
        command = calloc(command_len+1, sizeof(char));
        strncpy(command, str, command_len);

        printf("command is %s\n", command);
        for (str = p; str && *str == ' '; str++);
        ecore_list_prepend(parsed_msg, command);
      }
    else
      {
        return NULL;
      }
  }

  /* Get the params */
  while (str && *str != ':' && ECHOES_IS_NOT_SPACE(str))
  {
    int params_len;

    for (p = str; ECHOES_IS_NOT_SPACE(p); p++);
    if ((params_len = p - str))
      {
        params = calloc(params_len+1, sizeof(char));
        strncpy(params, str, params_len);

        printf("params is %s\n", params);
        ecore_list_append(parsed_msg, params);
        for (str = p; str && *str == ' '; str++);
      }
  }

  /* Get everything left until the \r\n */
  int trailing_len;
  p = msg + strlen(msg);
  if ((trailing_len = p - str))
    {
      trailing = calloc(trailing_len+1, sizeof(char));
      strncpy(trailing, str, trailing_len);
      printf("trailing is %s\n", trailing);
      ecore_list_append(parsed_msg, trailing);
    }

  return parsed_msg;
}

void
evas_list_pawned(Evas_List *list)
{
  Evas_List *l;

  for (l = list; l; l = l->next) 
    {
      free(l->data);
    }
  evas_list_free(list);
}

Echoes_Server *
echoes_server_add(Echoes *echoes, Ecore_Con_Server *server)
{
  Echoes_Server *es;

  es = calloc(1, sizeof(Echoes_Server));
  es->type = ECHOES_SERVER_TYPE_SERV;
  es->name = strdup("Coin");
  es->nick = calloc(BUF_MAX, sizeof(char));
  es->nick = strncpy(es->nick, IRC_NICK, 64);
  es->server = server;
  es->chans = ecore_list_new();
  ecore_list_set_free_cb(es->chans, echoes_server_free);
  ecore_hash_set(echoes->serv_hash, server, es);

  echoes_ui_server_add(echoes, es->name, es);

  return es;
}

void
echoes_server_del(Echoes *echoes, Echoes_Server *es)
{
  Echoes_Server *tmp;

  if (!es) return;
  echoes_ui_server_del(echoes, es);
  if (es->type == ECHOES_SERVER_TYPE_SERV)
    {
      ecore_list_destroy(es->chans);
      tmp = ecore_hash_remove(echoes->serv_hash, es->server);
      echoes_server_free(tmp);
    }
  else
    {
      ecore_hash_remove(echoes->chan_hash, es->name);
      ecore_list_find(es->parent->chans, ecore_direct_compare, es);
      ecore_list_remove_destroy(es->parent->chans);
    }
}

void
echoes_server_disconnect(void *data)
{
  Ecore_Con_Server *server;
  server = data;
  echoes_irc_quit(server, IRC_QUIT_MSG);
  ecore_con_server_del(server);
}

Echoes_Server *
echoes_server_get(Echoes *echoes)
{
  Echoes_Server *es;

  es = echoes_ui_server_get(echoes);
  return es;
}

void
echoes_server_free(void *data)
{
  Echoes_Server *es = data;
  if (es->type == ECHOES_SERVER_TYPE_SERV)
    {
      free(es->nick);
    }
  else if (es->type == ECHOES_SERVER_TYPE_CHAN)
    {
      echoes_irc_part(es->server, es->name, IRC_PART_MSG);
    }
  if (es->name) free(es->name);
  if (es->nicks) ecore_list_destroy(es->nicks);
  if (es->history) evas_list_pawned(es->history);
  free(es);
}

Echoes_Server *
echoes_chan_add(Echoes *echoes, const char *name, Echoes_Server *parent)
{
  Echoes_Server *es;

  if (!parent || !parent->server) return NULL;
  es = calloc(1, sizeof(Echoes_Server));
  es->type   = ECHOES_SERVER_TYPE_CHAN;
  es->name   = strdup(name);
  es->server = parent->server;
  es->nick   = parent->nick;
  es->parent = parent;
  ecore_list_append(parent->chans, es);

  ecore_hash_set(echoes->chan_hash, strdup(name), es);
  echoes_ui_server_add(echoes, es->name, es);
  return es;
}

Echoes_Server *
echoes_conv_add(Echoes *echoes, const char *name, Echoes_Server *parent)
{
  Echoes_Server *es;

  if (!parent || !parent->server) return NULL;
  es = calloc(1, sizeof(Echoes_Server));
  es->type   = ECHOES_SERVER_TYPE_CONV;
  es->name   = strdup(name);
  es->server = parent->server;
  es->nick   = parent->nick;
  es->parent = parent;
  ecore_list_append(parent->chans, es);

  ecore_hash_set(echoes->chan_hash, strdup(name), es);
  echoes_ui_server_add(echoes, es->name, es);
  return es;
}

void
echoes_nick_list_append(Echoes *echoes, Echoes_Server *es, const char *list)
{
  Ecore_List *nicks = NULL;
  char *nick;

  if (!es) return;
  if (!list)
    {
      if (es->nicks) ecore_list_destroy(es->nicks);
      es->nicks = NULL;
      echoes_ui_nick_fill(echoes, NULL);
      return;
    }

  if (es->nicks)
    nicks = es->nicks;
  else
    nicks = ecore_list_new();

  //TODO trouver mieux que ça c'est pas super
  char *str, *p;
  p = str = strdup(list);
  while ((nick = strsep(&str, " ")) && *nick != '\0')
    {
      ecore_list_append(nicks, strdup(nick));
    }
  ecore_list_set_free_cb(nicks, free);
  ecore_list_sort(nicks, _echoes_strcase_compare, 0);
  es->nicks = nicks;
  free(p);

  echoes_ui_nick_fill(echoes, nicks);
}

void
echoes_process_msg_out(Echoes *echoes, const char *msg)
{
  Ecore_List *words;
  Echoes_Server *es;
  int n_words;

  words = echoes_split_string(msg, " ");
  n_words = ecore_list_nodes(words);
  es = echoes_server_get(echoes);
  if (!words) return;
  if (*msg != '/' && es && es->type != ECHOES_SERVER_TYPE_SERV)
    {
       echoes_irc_privmsg(es->server, es->name, msg);
       echoes_ui_chan_msg(es, es->nick, (char *)msg);
    }
  else if (!strcasecmp(ecore_list_current(words), "/server"))
    {
      char *server;
      Ecore_Con_Server *con;

      server = ecore_list_goto_index(words, 1);

      if (ecore_list_nodes(words) < 3)
        con = echoes_irc_connect(echoes, server, 6667);
      else
        con = echoes_irc_connect(echoes, server, atoi(ecore_list_goto_index(words, 2)));
      if (con)
        {
          echoes_irc_nick(con, echoes->nick);
          echoes_irc_user(con, echoes->nick, server, IRC_USERNAME);
        }
    }
  else if (n_words >= 2 && !strcasecmp(ecore_list_current(words), "/join"))
    {
      char *chan, *key;

      chan = ecore_list_goto_index(words, 1);
      key  = ecore_list_goto_index(words, 2);
      if (es) echoes_irc_join(es->server, chan, key);
    }
  else if (n_words >= 2 && !strcasecmp(ecore_list_current(words), "/part"))
    {
      char *reason;

      reason = _echoes_strstr(msg+6, ecore_list_goto_index(words, 1));
      if (es) 
        {
          echoes_irc_part(es->server, es->name, reason);
          echoes_nick_list_append(echoes, es, NULL);
        }
    }
  else if (!strcasecmp(ecore_list_current(words), "/nick"))
    {
      char *nick;

      nick = ecore_list_goto_index(words, 1);
      if (es) echoes_irc_nick(es->server, nick);
    }
  else if (!strcasecmp(ecore_list_current(words), "/kick"))
    {
      char *nick, *reason;

      nick = ecore_list_goto_index(words, 1);
      reason = _echoes_strstr(msg+strlen(nick)+7, ecore_list_goto_index(words, 2));
      if (es) echoes_irc_kick(es->server, es->name, nick, reason);
    }
  else if (n_words >= 2 && !strcasecmp(ecore_list_current(words), "/mode"))
    {
      char *mode, *data;

      mode = ecore_list_goto_index(words, 1);
      data = ecore_list_goto_index(words, 2);
      //TODO Etendre le parsing selon la spec
      if (es) echoes_irc_mode(es->server, es->name, mode, data);
    }
  else if (!strcasecmp(ecore_list_current(words), "/topic"))
    {
      char *topic;

      topic = _echoes_strstr(msg+7, ecore_list_goto_index(words, 1));
      if (es) echoes_irc_topic(es->server, es->name, topic);
    }
  else if (!strcasecmp(ecore_list_current(words), "/names"))
    {
      if (es) echoes_irc_names(es->server, es->name);
      echoes_ui_names_msg(es);
    }
  //TODO /list
  else if (n_words >= 2 && !strcasecmp(ecore_list_current(words), "/invite"))
    {
      char *nick, *chan;

      nick = ecore_list_goto_index(words, 1);
      chan = ecore_list_goto_index(words, 2);
      if (es && chan)
        echoes_irc_invite(es->server, chan, nick);
      else if (es) 
        echoes_irc_invite(es->server, es->name, nick);
    }
  else if (!strcasecmp(ecore_list_current(words), "/quit") || !strcmp(ecore_list_current(words), ":q"))
    {
      char *reason;

      reason = _echoes_strstr(msg+5, ecore_list_goto_index(words, 1));
      if (es) echoes_irc_quit(es->server, reason);
      etk_main_quit();
    }
  else if (n_words >= 3 && !strcasecmp(ecore_list_current(words), "/msg"))
    {
      const char *nick, *str;
      nick = ecore_list_goto_index(words, 1);
      str  = _echoes_strstr(msg+strlen(nick)+6, ecore_list_goto_index(words, 2));
      if (es) echoes_irc_privmsg(es->server, nick, str);
    }
  else if (n_words >= 3 && !strcasecmp(ecore_list_current(words), "/notice"))
    {
      const char *nick, *str;
      nick = ecore_list_goto_index(words, 1);
      str = _echoes_strstr(msg+strlen(nick)+9, ecore_list_goto_index(words, 2));
      if (es) echoes_irc_notice(es->server, nick, str);
    }
  else if (n_words >= 3 && !strncasecmp(ecore_list_current(words), "/oper", 5))
    {
      char *nick, *pass;
      nick = ecore_list_goto_index(words, 1);
      pass = ecore_list_goto_index(words, 2);
      if (es) echoes_irc_oper(es->server, nick, pass);
    }
  else if (!strncasecmp(ecore_list_current(words), "/version", 8))
    {
      //TODO Peut etre utiliser du CTCP
      char *server;
      server = ecore_list_goto_index(words, 1);
      if (es) echoes_irc_version(es->server, server);
    }
  else if (!strncasecmp(ecore_list_current(words), "/stats", 6))
    {
      char *query;
      query = ecore_list_goto_index(words, 1);
      if (es) echoes_irc_stats(es->server, query);
    }
  else if (!strncasecmp(ecore_list_current(words), "/links", 6))
    {
      char *mask, *remote;
      mask = ecore_list_goto_index(words, 1);
      remote = ecore_list_goto_index(words, 2);
      if (es) echoes_irc_links(es->server, remote, mask);
    }
  else if (!strncasecmp(ecore_list_current(words), "/time", 5))
    {
      char *name;
      name = ecore_list_goto_index(words, 1);
      if (es) echoes_irc_time(es->server, name);
    }
  else if (!strncasecmp(ecore_list_current(words), "/info", 5))
    {
      char *name;
      name = ecore_list_goto_index(words, 1);
      if (es) echoes_irc_info(es->server, name);
    }
  else if (!strncasecmp(ecore_list_current(words), "/ping", 5))
    {
      char *name;
      name = ecore_list_goto_index(words, 1);
      if (es) echoes_irc_ping(es->server, name);
    }
  else if (!strncasecmp(ecore_list_current(words), "/who", 4))
    {
      char *name, *o;
      name = ecore_list_goto_index(words, 1);
      o    = ecore_list_goto_index(words, 2);
      if (es) echoes_irc_who(es->server, name, o);
    }
  else if (n_words >= 2 && !strncasecmp(ecore_list_current(words), "/whois", 6))
    {
      char *nickmask;
      nickmask = _echoes_strstr(msg+7, ecore_list_goto_index(words, 1));
      if (es) echoes_irc_whois(es->server, nickmask);
    }
  else if (n_words >= 2 && !strncasecmp(ecore_list_current(words), "/whowas", 6))
    {
      char *nick, *count, *name;
      nick = ecore_list_goto_index(words, 1);
      count = ecore_list_goto_index(words, 2);
      name = ecore_list_goto_index(words, 3);
      if (es) echoes_irc_whowas(es->server, nick, count, name);
    }
  else if (!strncasecmp(ecore_list_current(words), "/raw", 4))
    {
      char *cmd;
      cmd = _echoes_strstr(msg+5, ecore_list_goto_index(words, 1));
      if (es) echoes_irc_raw(es->server, cmd);
    }
  ecore_list_destroy(words);

  if (!es) return;
  /* Keep the msg in the history */
  es->history = evas_list_prepend(es->history, strdup(msg));
  if (evas_list_count(es->history) > 100)
    {
      Evas_List *l;
      l = evas_list_nth_list(es->history, 100);
      es->history = evas_list_remove_list(es->history, l);
      evas_list_pawned(l);
    }
  es->cursor = NULL;
}

void
echoes_process_msg_in(Echoes *echoes, Ecore_Con_Server *in_serv, const char *msg)
{
  Ecore_List *lines;
  char *line;

  lines = echoes_split_string(msg, "\r\n");
  while ((line = ecore_list_next(lines)))
    {
      Ecore_List *words;
      char *rpl_word;
      int rpl_code = 0;
      int n_words;

      printf("%s\n", line);

      words = echoes_parse_message(line);
      if (!words) continue;
      n_words = ecore_list_nodes(words);
      rpl_word = ecore_list_goto_first(words);
      if (n_words > 1)
        rpl_code = atoi(rpl_word);

      if (n_words >= 3 && !strcmp(rpl_word, "PRIVMSG"))
        {
          Echoes_Server *es;
          char *nick, *chan, *txt;

          nick = malloc(BUF_MAX);
          _echoes_parse_nick_get(nick, ecore_list_goto_index(words, 1));
          chan = ecore_list_goto_index(words, 2);
          //TODO Parser CTCP http://www.invlogic.com/irc/ctcp.html

          es = ecore_hash_get(echoes->serv_hash, in_serv);
          if (!strcasecmp(chan, es->nick))
            {
              if (!ecore_hash_get(echoes->chan_hash, nick))
                es = echoes_conv_add(echoes, nick, es);
              else
                es = ecore_hash_get(echoes->chan_hash, nick);
            }
          else
            {
              es = ecore_hash_get(echoes->chan_hash, chan);
            }

          txt  = ecore_list_goto_index(words, 3);
          echoes_ui_chan_msg(es, nick, txt+1);
          free(nick);
        }
      else if (n_words > 1 && !strcmp(rpl_word, "PING"))
        {
          char *server;

          server = ecore_list_goto_index(words, 1);
          echoes_irc_pong(in_serv, server+1);
        }
      else if (n_words > 2 && !strcmp(rpl_word, "JOIN"))
        {
          char *nick, *user, *chan;
          Echoes_Server *es;

          es = ecore_hash_get(echoes->serv_hash, in_serv);
          nick = malloc(BUF_MAX);
          user = _echoes_parse_nick_get(nick, ecore_list_goto_index(words, 1));
          chan = ecore_list_goto_index(words, 2);
          if (!strcmp(nick, es->nick))
            {
              es = echoes_chan_add(echoes, chan+1, es);
            }
          else
            {
              es = ecore_hash_get(echoes->chan_hash, chan+1);
              echoes_nick_add(echoes, nick, chan+1);
            }
          echoes_ui_join_msg(es, nick, user);
          free(nick);
        }
      else if (n_words > 2 && !strcmp(rpl_word, "PART"))
        {
          Echoes_Server *es;
          char  *nick, *user, *chan, *part_msg;

          nick = malloc(BUF_MAX);
          user = _echoes_parse_nick_get(nick, ecore_list_goto_index(words, 1));
          chan = ecore_list_goto_index(words, 2);
          part_msg = ecore_list_goto_index(words, 3);

          es = ecore_hash_get(echoes->chan_hash, chan);
          echoes_ui_part_msg(es, nick, user, part_msg+1);
          echoes_nick_del(echoes, es, nick);
          free(nick);
        }
      else if (n_words >= 3 && !strcmp(rpl_word, "QUIT"))
        {
          Echoes_Server *es;
          char  *nick, *user, *quit_msg;

          es = ecore_hash_get(echoes->serv_hash, in_serv);
          nick = malloc(BUF_MAX);
          user = _echoes_parse_nick_get(nick, ecore_list_goto_index(words, 1));
          quit_msg = ecore_list_goto_index(words, 2);

          echoes_nick_quit(echoes, es, nick, quit_msg+1);
          free(nick);
        }
      else if (n_words > 4 && !strcmp(rpl_word, "KICK"))
        {
          Echoes_Server *es;
          char *nick, *op, *chan, *msg;

          op = malloc(BUF_MAX);
          _echoes_parse_nick_get(op, ecore_list_goto_index(words, 1));
          chan = ecore_list_goto_index(words, 2);
          nick = ecore_list_goto_index(words, 3);
          msg  = ecore_list_goto_index(words, 4);

          es = ecore_hash_get(echoes->chan_hash, chan);
          echoes_ui_kick_msg(es, op, nick, msg+1);
          echoes_nick_del(echoes, es, nick);
          free(op);
        }
      else if (n_words >= 3 && !strcmp(rpl_word, "NICK"))
        {
          Echoes_Server *es;
          char *old_nick, *new_nick;

          es = ecore_hash_get(echoes->serv_hash, in_serv);
          old_nick = malloc(BUF_MAX);
          _echoes_parse_nick_get(old_nick, ecore_list_goto_index(words, 1));
          new_nick = ecore_list_goto_index(words, 2);

          echoes_nick_switch(echoes, es, old_nick, new_nick+1);
          free(old_nick);
        }
      else if (n_words >= 3 && !strcmp(rpl_word, "MODE"))
        {
          Echoes_Server *es;
          char *src, *dest, *mode, *nick, *new_nick = NULL;

          src = malloc(BUF_MAX);
          _echoes_parse_nick_get(src, ecore_list_goto_index(words, 1));
          dest = ecore_list_goto_index(words, 2);
          mode = ecore_list_goto_index(words, 3);
          nick = ecore_list_goto_index(words, 4);
          if (*dest == '#' || *dest == '!' || *dest == '&') 
            es = ecore_hash_get(echoes->chan_hash, dest);
          else
            es = ecore_hash_get(echoes->serv_hash, in_serv);
          new_nick = echoes_mode_set(es, mode, nick);
          if (new_nick) echoes_ui_nick_switch(echoes, nick, new_nick);
          echoes_ui_mode_msg(es, src, mode, nick);
          free(src);
        }
      else if (n_words >= 4 && !strcmp(rpl_word, "INVITE"))
        {
          Echoes_Server *es;
          char *op, *nick, *chan;

          op = malloc(BUF_MAX);
          _echoes_parse_nick_get(op, ecore_list_goto_index(words, 1));
          nick = ecore_list_goto_index(words, 2);
          chan = ecore_list_goto_index(words, 3);
          
          es = ecore_hash_get(echoes->chan_hash, chan+1);
          if (!es) es = ecore_hash_get(echoes->serv_hash, in_serv);

          echoes_ui_invite_msg(es, chan+1, op, nick);
          free(op);
        }
      else if (n_words >= 3 && rpl_code == 1)
        {
          Echoes_Server *es;
          char *server_name, *msg;

          server_name = ecore_list_goto_index(words, 1);
          msg = ecore_list_goto_index(words, 3);

          es = ecore_hash_get(echoes->serv_hash, in_serv);
          if (es->name) free(es->name);
          es->name = strdup(server_name+1);
          echoes_ui_server_name_update(es);
          if (es) echoes_ui_serv_msg(es, msg+1);
        }
      else if (rpl_code == RPL_TOPIC)
        {
          Echoes_Server *es;
          char *chan;
          char *topic;

          chan = ecore_list_goto_index(words, 3);
          topic = ecore_list_goto_index(words, 4);
          es = ecore_hash_get(echoes->chan_hash, chan);
          echoes_ui_topic_set(es, topic+1);
        }
      else if (n_words > 5 && rpl_code == RPL_ENDOFTOPIC)
        {
          Echoes_Server *es;
          char  *chan;
          char  *nick;
          time_t date;

          chan = ecore_list_goto_index(words, 3);
          nick = ecore_list_goto_index(words, 4);
          date = strtoul(ecore_list_goto_index(words, 5), NULL, 10);
          es = ecore_hash_get(echoes->chan_hash, chan);
          echoes_ui_topic_end_set(es, nick, date);
        }
      else if (n_words >= 6 && rpl_code == RPL_NAMREPLY)
        {
          Echoes_Server *es;
          char *chan, *nicks;

          chan = ecore_list_goto_index(words, 4);
          nicks = ecore_list_goto_index(words, 5); 
          es = ecore_hash_get(echoes->chan_hash, chan);
          echoes_nick_list_append(echoes, es, nicks+1);
        }
      else if (n_words > 3)
        {
          char *source, *msg;
          Echoes_Server *es;

          msg = ecore_list_goto_index(words, 3);
          es = ecore_hash_get(echoes->serv_hash, in_serv);
          if (es) echoes_ui_serv_msg(es, msg+1);
        }
      ecore_list_destroy(words);
    }
  ecore_list_destroy(lines);
}

static char *
_echoes_parse_nick_get(char *dest, char *src)
{
  char *str;
  int i;

  str = src;
  for (++str, i = 0; str && *str != '!' && i < BUF_MAX; str++, i++)
    {
      dest[i] = *str;
    }
  dest[i] = '\0';
  if (*str == '!')
    return ++str;
  else 
    return NULL;
}

static char *
_echoes_strstr(const char *haystack, const char *needle)
{
  if (!haystack || !needle) 
    return NULL;
  else
    return strstr(haystack, needle);
}

void
echoes_history_prev(Echoes *echoes)
{
  Echoes_Server *es;

  es = echoes_server_get(echoes);
  if (!es) return;

  if (!es->cursor)
    es->cursor = es->history;
  else
    es->cursor = evas_list_next(es->cursor);
  echoes_ui_user_msg_set(echoes, (const char *)evas_list_data(es->cursor));
}

void
echoes_history_next(Echoes *echoes)
{
  Echoes_Server *es;

  es = echoes_server_get(echoes);
  if (!es) return;

  es->cursor = evas_list_prev(es->cursor);
  echoes_ui_user_msg_set(echoes, (const char *)evas_list_data(es->cursor));
}

void
echoes_nick_add(Echoes *echoes, const char *nick, const char *chan)
{
  Echoes_Server *es;
  char *str;

  if (!chan) return;

  es = ecore_hash_get(echoes->chan_hash, chan);
  if (!es || !es->nicks) return;
  if (!strcasecmp(nick, es->nick)) return;

  ecore_list_goto_first(es->nicks);
  ecore_list_append(es->nicks, strdup(nick));
  ecore_list_sort(es->nicks, _echoes_strcase_compare, 0);

  es = echoes_server_get(echoes);
  if (es && !strcasecmp(es->name, chan))
    echoes_ui_nick_fill(echoes, es->nicks);
}

void
echoes_nick_del(Echoes *echoes, Echoes_Server *es, const char *nick)
{
  char *str;

  if (!es || !es->nicks) return;
  str = ecore_list_find(es->nicks, _echoes_nick_compare, nick);
  if (str)
    ecore_list_remove_destroy(es->nicks);
  else
    ecore_list_goto_first(es->nicks);

  if (es == echoes_server_get(echoes))
    echoes_ui_nick_del(echoes, nick);
}

void
echoes_nick_quit(Echoes *echoes, Echoes_Server *es, const char *nick, const char *quit_msg)
{
  struct {
    const char *nick;
    const char *quit_msg;
  } qdata;
  qdata.nick = nick;
  qdata.quit_msg = quit_msg;
  ecore_list_for_each(es->chans, _echoes_nick_quit_cb, &qdata);
  echoes_ui_nick_del(echoes, nick);
}

void
echoes_nick_switch(Echoes *echoes, Echoes_Server *es, const char *old_nick, const char *new_nick)
{
  struct {
    const char *old_nick;
    const char *new_nick;
  } snick;

  if (!es) return;

  snick.old_nick = old_nick;
  snick.new_nick = new_nick;
  if (!strcmp(es->nick, old_nick))
    {
      Echoes_Server *current_es;

      current_es = echoes_server_get(echoes);
      strncpy(es->nick, new_nick, strlen(new_nick)+1);
      if (es == current_es->parent || es == current_es) 
        echoes_ui_nick_set(current_es, new_nick);
    }

  ecore_list_for_each(es->chans, _echoes_nick_switch_cb, &snick);
  echoes_ui_nick_switch(echoes, old_nick, new_nick);
}

Ecore_List *
echoes_nick_completion(Echoes_Server *es, const char *starts_with)
{
  Ecore_List *nicks = NULL;
  char *nick;
  int   len;
  int   ret;

  if (!es || !es->nicks || !starts_with) return NULL;
  len = strlen(starts_with);
  nicks = ecore_list_new();
  ecore_list_init(nicks);
//  ecore_list_set_free_cb(nicks, free);

  ecore_list_goto_first(es->nicks);
  while ((nick = ecore_list_next(es->nicks)))
    {
      if ((ret = strncmp(starts_with, nick, len)) == 0)
        ecore_list_append(nicks, nick);
      else if (ret < 0)
        break;
    }

  if (ecore_list_nodes(nicks) == 0)
    {
      ecore_list_destroy(nicks);
      return NULL;
    }
  else
    {
      return nicks;
    }
}

char *
echoes_list_to_str(Ecore_List *list)
{
  char *strlist;

  if (!list) return NULL;
  ecore_list_goto_first(list);
  strlist = calloc(2, sizeof(char));
  *strlist = ' ';
  ecore_list_for_each(list, _echoes_strlist_concat_cb, &strlist);
  return strlist;
}

char *
echoes_mode_set(Echoes_Server *es, const char *mode, const char *nick)
{
  char *str;
  const char *flag;
  char *new_nick;
  if (!es || !es->nicks || !mode || !nick) return NULL;

  flag = mode;
  str = ecore_list_find(es->nicks, _echoes_nick_compare, nick);
  if (str)
    {
      char buf[BUF_MAX];
      if (*flag == '+')
        {
          for (flag++; *flag != '\0'; flag++)
            {
              if (*flag == 'o' && *str != '@')
                {
                  if (*str == '+')
                    strcpy(buf, str+1);
                  else
                    strcpy(buf, str);
                  str = strncpy(str, "@", 2);
                }
              else if (*flag == 'v' && *str != '+' && *str != '@')
                {
                  strcpy(buf, str);
                  str = strncpy(str, "+", 2);
                }
            }
           str = strcat(str, buf);
        }
      else
        {
          char *buf = NULL;
          for (flag++; *flag != '\0'; flag++)
            {
              if (*flag == 'o' && *str == '@')
                {
                  buf = str+1;
                }
              else if (*flag == 'v' && *str == '+')
                {
                  buf = str+1;
                }
            }
          if (buf) strcpy(str, buf);
        }
    }
  return str;
}

static void
_echoes_nick_switch_cb(void *data, void *udata)
{
  Echoes_Server *es;
  const char *str = NULL;
  struct {
    const char *old_nick;
    const char *new_nick;
  } *snick;

  es = data;
  snick = udata;
  if (!es->nicks) return;

  str = ecore_list_find(es->nicks, _echoes_nick_compare, snick->old_nick);
  if (str)
    {
      printf("%s\n", str);
      char *buf;

      buf = calloc(strlen(snick->new_nick)+2, sizeof(char));
      if (*str == '@' || *str == '&' || *str == '+') strncpy(buf, str, 1);
      buf = strcat(buf, snick->new_nick);
      ecore_list_remove_destroy(es->nicks);
      ecore_list_append(es->nicks, buf);
      echoes_ui_nick_msg(es, snick->old_nick, snick->new_nick);
    }
  else
    ecore_list_goto_first(es->nicks);

  return;
}

static int
_echoes_nick_compare(const void *d1, const void *d2)
{
  const char *str1, *str2;
  str1 = d1;
  str2 = d2;
  if (*str1 == '@' || *str1 == '&' || *str1 == '+' ) str1++;
  if (*str2 == '@' || *str2 == '&' || *str2 == '+' ) str2++;
  return strcasecmp(str1, str2);
}

static int
_echoes_strcase_compare(const void *d1, const void *d2)
{
  const char *c1, *c2;
  c1 = d1;
  c2 = d2;
  return strcasecmp(c1, c2);
}

static void
_echoes_nick_quit_cb(void *data, void *udata)
{
  Echoes_Server *es;
  const char *str = NULL;
  struct {
    const char *nick;
    const char *quit_msg;
  } *qdata;

  es = data;
  qdata = udata;
  if (!es->nicks) return;

  str = ecore_list_find(es->nicks, _echoes_strcase_compare, qdata->nick);
  if (str)
    {
      ecore_list_remove_destroy(es->nicks);
      echoes_ui_quit_msg(es, qdata->nick, qdata->quit_msg);
    }
  else
    ecore_list_goto_first(es->nicks);

  return;
}

static void
_echoes_strlist_concat_cb(void *data, void *fdata)
{
  char *str;
  char **strlist;

  str = data;
  strlist = fdata;
  *strlist = realloc(*strlist, strlen(*strlist)+strlen(str)+2);
  *strlist = strcat(*strlist, str);
  *strlist = strcat(*strlist, " ");
}
