#ifndef ECHOES_H
#define ECHOES_H

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>

#include <Etk.h>
#include <Enhance.h>
#include <Evas.h>
#include <Ecore.h>
#include <Ecore_Con.h>
#include <Ecore_Data.h>

#include "config.h"

#define ecore_hash_remove_destroy(hash, key) hash->free_value(ecore_hash_remove(hash, key))

typedef enum _Echoes_Server_Type Echoes_Server_Type;
enum _Echoes_Server_Type
{
  ECHOES_SERVER_TYPE_SERV = 1 << 0,
  ECHOES_SERVER_TYPE_CHAN = 1 << 1,
  ECHOES_SERVER_TYPE_CONV = 1 << 2
};

typedef enum _Echoes_Server_Status Echoes_Server_Status;
enum _Echoes_Server_Status
{
  ECHOES_STATUS_DEAD = 1 << 0,
  ECHOES_STATUS_NONE = 1 << 1,
  ECHOES_STATUS_NOTE = 1 << 2,
  ECHOES_STATUS_TALK = 1 << 3
};

typedef struct _Echoes_Connection Echoes_Connection;
struct _Echoes_Connection
{
  Ecore_Con_Server *server;
  Ecore_Event_Handler *add_hdl;
  Ecore_Event_Handler *del_hdl;
  Ecore_Event_Handler *data_hdl;
};

typedef struct _Echoes_Server Echoes_Server;
struct _Echoes_Server 
{
  char *nick;
  char *name;

  Echoes_Server_Type type;
  Echoes_Server_Status status;
  
  Etk_Widget *frame;
  Etk_Widget *scrolled_view;
  Etk_Widget *textview;
  Etk_Widget *msg_entry;
  Etk_Widget *nick_button;
  Etk_Tree_Row *row;
  Etk_Notebook_Page *page;

  Ecore_Con_Server *server;

  Ecore_List *nicks; //Used only by channels
  Ecore_List *chans; //Used only by servers
  Echoes_Server *parent; //Used only by channels

  Evas_List *history;
  Evas_List *cursor;
};

typedef struct _Echoes Echoes;
struct _Echoes
{
  char *nick;
  Ecore_Hash *chan_hash;
  Ecore_Hash *serv_hash;

  Etk_Widget *server_list;
  Etk_Widget *nick_list;
  Etk_Widget *notebook;
  Etk_Widget *logo;
  Etk_Widget *nick_button;
  Etk_Widget *msg_entry;

  char *incomplete;
  Enhance *en;
};

#include "echoes_irc.h"
#include "echoes_ui.h"
#include "echoes_callbacks.h"

#define BUF_MAX 512
#define IRC_NICK "Echoes_svn"
#define IRC_USERNAME "Echoes_User"
#define IRC_QUIT_MSG "Echoes fade away."
#define IRC_PART_MSG "Echoes fade away."

#define IRC_SERV_MSG         "<font size=11 color=#df8989>%s </font>"
#define IRC_CHAN_ACTION      "<font size=11 color=#723ac6>[%s] <b>*%s </b>%s </font>"
#define IRC_CHAN_YOU_MSG     "<font size=11 color=#e86739>[%s] <b>%s:</b> %s </font>" //dernier espace : bug utf8 ; mais la je comprends pourquoi, ca devrait pas.
#define IRC_CHAN_SOME_MSG    "<font size=11 color=#1a15de>[%s] <b>%s:</b> %s </font>" //dernier espace : bug utf8 ; mais la je comprends pourquoi, ca devrait pas.
#define IRC_CHAN_SOME_HL_MSG "<style effect=glow color1=#fda61e><font size=11 color=#f3de31>[%s] <b>%s:</b> %s </font></style>" // dernier espace : bug utf8
#define IRC_CHAN_TOPIC       "<font size=11 color=#723ac6>Topic for %s is %s </font>" // dernier espace : bug utf8
#define IRC_CHAN_TOPIC_END   "<font size=11 color=#723ac6>Topic set by %s on %s </font>"
#define IRC_YOU_JOIN         "<font size=11 color=#8f2cbd>Now talking on %s</font>"
#define IRC_SOME_JOIN        "<font size=11 color=#9ed965>--<font>&gt;</font> %s (%s) has joined </font>"
#define IRC_YOU_PART         "<font size=11 color=#8f2cbd>You have left channel %s (%s)</font>"
#define IRC_SOME_PART        "<font size=11 color=#d54758>&lt;-- %s has left %s (%s)</font>"
#define IRC_YOU_KICK         "<font size=11 color=#ff0000>You have been kicked from %s by %s because %s </font>" // dernier espace : bug utf8
#define IRC_SOME_KICK        "<font size=11 color=#2d9cbc>%s has been kicked from %s by %s because %s </font>"  // dernier espace : bug utf8
#define IRC_YOU_NICK         "<font size=11 color=#8f2cbd>You are now known as %s</font>"
#define IRC_SOME_NICK        "<font size=11 color=#9ed965>%s is now known as %s</font>"
#define IRC_NAMES            "<font size=11 color=#c3d5b8>Users on %s: %s</font>"
#define IRC_YOU_ARE_INVITE   "<font size=11 color=#723ac6>You are invited on %s by %s </font>"
#define IRC_YOU_MAKE_INVITE  "<font size=11 color=#723ac6>You are inviting %s on %s </font>"
#define IRC_YOU_QUIT         "<font size=11 color=#8f2cbd>You have quit (%s)</font>"
#define IRC_SOME_QUIT        "<font size=11 color=#d54758>&lt;-- %s has quit (%s)</font>"
#define IRC_MODE_MSG         "<font size=11 color=#723ac6>%s sets mode %s on %s</font>"

void echoes_defaults_set(Echoes *echoes);
Ecore_List *echoes_split_string(const char *str, const char *delim);
void evas_list_pawned(Evas_List *list);

Echoes_Server *echoes_server_add(Echoes *echoes, Ecore_Con_Server *server);
void           echoes_server_del(Echoes *echoes, Echoes_Server *es);
Echoes_Server *echoes_server_get(Echoes *echoes);
void           echoes_server_free(void *data);
void echoes_server_disconnect(void *data);

Echoes_Server *echoes_chan_add(Echoes *echoes, const char *name, Echoes_Server *parent);
Echoes_Server *echoes_conv_add(Echoes *echoes, const char *name, Echoes_Server *parent);

void echoes_nick_add(Echoes *echoes, const char *nick, const char *chan);
void echoes_nick_del(Echoes *echoes, Echoes_Server *es, const char *nick);
void echoes_nick_switch(Echoes *echoes, Echoes_Server *es, const char *old_nick, const char *new_nick);
void echoes_nick_quit(Echoes *echoes, Echoes_Server *es, const char *nick, const char *quit_msg);
Ecore_List *echoes_nick_completion(Echoes_Server *es, const char *starts_with);
void echoes_nick_list_append(Echoes *echoes, Echoes_Server *es, const char *list);

Ecore_List *echoes_parse_message(const char *msg);
void echoes_process_msg_out(Echoes *echoes, const char *msg);
void echoes_process_msg_in(Echoes *echoes, Ecore_Con_Server *in_serv, const char *msg);

void echoes_history_prev(Echoes *echoes);
void echoes_history_next(Echoes *echoes);

char *echoes_list_to_str(Ecore_List *list);
char *echoes_mode_set(Echoes_Server *es, const char *mode, const char *nick);

#endif /* ECHOES_H */
