#include "echoes.h"
#include "echoes_callbacks.h"

Etk_Bool
on_window_quit(Etk_Object *obj, void *data)
{
  // FIXME méthode barbare
  etk_main_quit();
  return ETK_TRUE;
}

void
on_msg_entry_key_down_event(Etk_Object *obj, Etk_Event_Key_Down *ev, void *data)
{
  Echoes *echoes = data;

  if (!strcmp(ev->key, "Return"))
    {
      const char *msg;

      msg = echoes_ui_user_msg_get(echoes);
      echoes_process_msg_out(echoes, msg);
      echoes_ui_user_msg_clear(echoes);
    }
  else if (!strcmp(ev->key, "Up"))
    {
      echoes_history_prev(echoes);
    }
  else if (!strcmp(ev->key, "Down"))
    {
      echoes_history_next(echoes);
    }
  else if (ev->modifiers == ETK_MODIFIER_CTRL && !strcasecmp(ev->key, "w"))
    {
      Echoes_Server *es;
      es = echoes_server_get(echoes);
      if (es) echoes_server_del(echoes, es);
    }
  else if (!strcmp(ev->keyname, "Tab"))
    {
      Echoes_Server *es;
      es = echoes_server_get(echoes);
      if (es)
        {
          Ecore_List *completion;
          char *possible_nicks, *txt;

          completion = echoes_nick_completion(es, echoes_ui_user_msg_get(echoes));
          if (!completion) return;
          if (ecore_list_nodes(completion) == 1)
            {
              ecore_list_goto_first(completion);
              echoes_ui_user_msg_clear(echoes);
              asprintf(&txt, "%s, ", (char *)ecore_list_current(completion));
              echoes_ui_user_msg_set(echoes, txt);
            }
          else
            {
              possible_nicks = echoes_list_to_str(completion);
              asprintf(&txt, "\n%s", possible_nicks);
              echoes_ui_direct_write_txt(es, txt);
              free(possible_nicks);
            }
          ecore_list_destroy(completion);
          free(txt);
        }
      etk_signal_stop();
    }
}

int
on_server_event_add(void *data, int type, void *event)
{
  Echoes *echoes;
  Ecore_Con_Event_Server_Add *ev;

  echoes = data;
  ev = event;

  echoes_server_add(echoes, ev->server);

  return 0;
}

int
on_server_event_del(void *data, int type, void *event)
{
  Echoes *echoes;
  Ecore_Con_Event_Server_Data *ev;

  echoes = data;
  ev   = event;
  return 0;
}

int
on_server_event_data(void *data, int type, void *event)
{
  Echoes *echoes;
  Ecore_Con_Event_Server_Data *ev;
  char *msg;

  echoes = data;
  ev = event;
  
  msg = calloc(ev->size+1, sizeof(char));
  strncpy(msg, ev->data, ev->size);
  if (echoes->incomplete)
    {
      char *str;
      str = msg;
      asprintf(&msg, "%s%s", echoes->incomplete, str);
      free(echoes->incomplete);
      free(str);
      echoes->incomplete = NULL;
    }

  if (msg[strlen(msg)-1] != '\n')
    {
      char *incomplete;
      incomplete = strrchr(msg, '\n'); 
      if (incomplete)
        {
          incomplete++;
          if (*incomplete == ':')
            {
              echoes->incomplete = strdup(incomplete);
              incomplete = memset(incomplete, 0, strlen(incomplete));
            }
        }
      else
        {
          free(msg);
          return 0;
        }
    }
  echoes_process_msg_in(echoes, ev->server, msg);
  free(msg);

  return 0;
}

void
on_server_list_row_select_event(Etk_Object *obj, Etk_Tree_Row *row, void *data)
{
  Echoes *echoes;
  Etk_Tree *tree;
  Etk_Tree_Col *col;
  Echoes_Server *es;

  echoes = data;
  tree = ETK_TREE(obj);

  es = etk_tree_row_data_get(row);
  if (!es) return;

  echoes_ui_server_show(es);
  echoes_ui_nick_fill(echoes, es->nicks);
  echoes_ui_nick_set(es, es->nick);
  echoes_ui_server_tree_notif(es, ECHOES_STATUS_NONE);
}

void
on_nick_list_row_select_event(Etk_Object *obj, Etk_Tree_Row *row, void *data)
{
  Echoes *echoes;
  Etk_Tree *tree;
  Etk_Tree_Col *col;

  echoes = data;
  tree = ETK_TREE(obj);
  col = etk_tree_nth_col_get(tree, 0);
}

void
on_nick_button_clicked(Etk_Object *obj, void *data)
{
  Echoes *echoes = data;
  Ecore_Con_Server *con;

  con = echoes_irc_connect(echoes, "irc.freenode.net", 6667);
  echoes_irc_nick(con, echoes->nick);
  echoes_irc_user(con, echoes->nick, "irc.freenode.net", IRC_USERNAME);
}
