#include "echoes.h"
#include "echoes_irc.h"

Ecore_Con_Server *
echoes_irc_connect(Echoes *echoes, const char *server, int port)
{
  //Echoes_Server *es;
  Ecore_Con_Server *con;

  //es = evas_hash_find(echoes->hash, server);
  con = ecore_con_server_connect(ECORE_CON_REMOTE_SYSTEM,
                                 server, port, NULL);
  if (!con)
    {
      fprintf(stderr, "Failed to open connection on server\n");
      return NULL;
    }

  ecore_event_handler_add(ECORE_CON_EVENT_SERVER_ADD,
                          on_server_event_add, echoes);
  ecore_event_handler_add(ECORE_CON_EVENT_SERVER_ADD,
                          on_server_event_del, echoes);
  ecore_event_handler_add(ECORE_CON_EVENT_SERVER_DATA,
                          on_server_event_data, echoes);

  //TODO c'est necessaire ?
  //echoes->connections = evas_list_append(echoes->connections, es);
  //es->server = con;
  return con;
}

void 
echoes_irc_nick(Ecore_Con_Server *server, const char *nick)
{
  int len=0;
  char out[BUF_MAX];

  if (!server) return;
  len = snprintf(out, sizeof(out), "NICK :%s\n", nick);
  ecore_con_server_send(server, out, len);
}

void 
echoes_irc_user(Ecore_Con_Server *server, const char *uname, 
              const char *sname, const char *rname)
{
  char host[64];
  int len=0;
  char out[BUF_MAX];

  if (!server) return;
  gethostname(host, 64);
  len = snprintf(out, sizeof(out), "USER %s %s %s :%s\r\n", 
                 uname, host, sname, rname);
  ecore_con_server_send(server, out, len);
}

void 
echoes_irc_join(Ecore_Con_Server *server, const char *channel, const char *key)
{
  int len=0;
  char out[BUF_MAX];

  if (!server) return;
  if (!key)
    len = snprintf(out, sizeof(out), "JOIN %s\n", channel);
  else
    len = snprintf(out, sizeof(out), "JOIN %s %s\n", channel, key);
  ecore_con_server_send(server, out, len);
}

void 
echoes_irc_part(Ecore_Con_Server *server, const char *channel, const char *reason)
{
  int len=0;
  char out[BUF_MAX];

  if (!server) return;
  if (!reason)
    len = snprintf(out, sizeof(out), "PART %s\n", channel);
  else
    len = snprintf(out, sizeof(out), "PART %s %s\n", channel, reason);
  ecore_con_server_send(server, out, len);
}

void 
echoes_irc_privmsg(Ecore_Con_Server *server, const char *dest, const char *msg)
{
  int len=0;
  char out[BUF_MAX];

  if (!server) return;
  len = snprintf(out, sizeof(out), "PRIVMSG %s :%s\n", dest, msg);
  ecore_con_server_send(server, out, len);
}

void 
echoes_irc_notice(Ecore_Con_Server *server, const char *dest, const char *msg)
{
  int len=0;
  char out[BUF_MAX];

  if (!server) return;
  len = snprintf(out, sizeof(out), "PRIVMSG %s :%s\n", dest, msg);
  ecore_con_server_send(server, out, len);
}

void 
echoes_irc_raw(Ecore_Con_Server *server,const char *msg)
{
  int len=0;
  char out[BUF_MAX];

  if (!server) return;
  len = snprintf(out, sizeof(out), "%s", msg);
  ecore_con_server_send(server, out, len);
}

void
echoes_irc_ping(Ecore_Con_Server *server, char *dest)
{
  int len=0;
  char out[BUF_MAX];

  if (!server) return;
  len = snprintf(out, sizeof(out), "PING %s\n", dest);
  ecore_con_server_send(server, out, len);
}

void
echoes_irc_pong(Ecore_Con_Server *server, char *dest)
{
  int len=0;
  char out[BUF_MAX];

  if (!server) return;
  len = snprintf(out, sizeof(out), "PONG %s\n", dest);
  ecore_con_server_send(server, out, len);
}

void 
echoes_irc_kick(Ecore_Con_Server *server, const char *chan, 
              const char *nick, const char *msg)
{
  int len=0;
  char out[BUF_MAX];

  if (!server) return;
  if (msg)
    len = snprintf(out, sizeof(out), "KICK %s %s :%s\n", chan, nick, msg);
  else
    len = snprintf(out, sizeof(out), "KICK %s %s\n", chan, nick);
  ecore_con_server_send(server, out, len);
}

void 
echoes_irc_mode(Ecore_Con_Server *server, const char *chan, 
              const char *mode, const char *data)
{
  int len=0;
  char out[BUF_MAX];

  if (!server) return;
  if (data)
    len = snprintf(out, sizeof(out), "MODE %s %s %s\n", chan, mode, data);
  else
    len = snprintf(out, sizeof(out), "MODE %s %s\n", chan, mode);
  ecore_con_server_send(server, out, len);
}

void
echoes_irc_quit(Ecore_Con_Server *server, const char *msg)
{
  int len=0;
  char out[BUF_MAX];

  if (!server) return;
  if (msg)
    len = snprintf(out, sizeof(out), "QUIT :%s\n", msg);
  else
    len = snprintf(out, sizeof(out), "QUIT\n");
  ecore_con_server_send(server, out, len);
  ecore_con_server_flush(server);
}

void
echoes_irc_oper(Ecore_Con_Server *server, const char *nick, const char *pass)
{
  int len=0;
  char out[BUF_MAX];

  if (!server) return;
  len = snprintf(out, sizeof(out), "OPER %s %s\n", nick, pass);
  ecore_con_server_send(server, out, len);
}

void
echoes_irc_topic(Ecore_Con_Server *server, const char *chan, const char *topic)
{
  int len=0;
  char out[BUF_MAX];

  if (!server) return;
  if (topic)
    len = snprintf(out, sizeof(out), "TOPIC %s :%s\n", chan, topic);
  else
    len = snprintf(out, sizeof(out), "TOPIC %s\n", chan);
  ecore_con_server_send(server, out, len);
}

void
echoes_irc_names(Ecore_Con_Server *server, const char *chan)
{
  int len=0;
  char out[BUF_MAX];

  if (!server) return;
  len = snprintf(out, sizeof(out), "NAMES %s\n", chan);
  ecore_con_server_send(server, out, len);
}

void
echoes_irc_invite(Ecore_Con_Server *server, const char *chan, const char *nick)
{
  int len=0;
  char out[BUF_MAX];

  if (!server) return;
  len = snprintf(out, sizeof(out), "INVITE %s %s\n", nick, chan);
  ecore_con_server_send(server, out, len);
}

void  
echoes_irc_version(Ecore_Con_Server *server, const char *name)
{
  int len=0;
  char out[BUF_MAX];

  if (!server) return;
  if (name)
    len = snprintf(out, sizeof(out), "VERSION %s\n", name);
  else
    len = snprintf(out, sizeof(out), "VERSION\n");
  ecore_con_server_send(server, out, len);
}

void  
echoes_irc_stats(Ecore_Con_Server *server, const char *query)
{
  int len=0;
  char out[BUF_MAX];

  if (!server) return;
  if (query)
    len = snprintf(out, sizeof(out), "STATS %s\n", query);
  else
    len = snprintf(out, sizeof(out), "STATS\n");
  ecore_con_server_send(server, out, len);
}

void  
echoes_irc_links(Ecore_Con_Server *server, const char *remote, const char *mask)
{
  int len=0;
  char out[BUF_MAX];

  if (!server) return;
  if (remote && mask)
    len = snprintf(out, sizeof(out), "LINKS %s %s\n", remote, mask);
  else if (mask)
    len = snprintf(out, sizeof(out), "LINKS %s\n", mask);
  else
    len = snprintf(out, sizeof(out), "LINKS\n");
  ecore_con_server_send(server, out, len);
}

void  
echoes_irc_time(Ecore_Con_Server *server, const char *name)
{
  int len=0;
  char out[BUF_MAX];

  if (!server) return;
  if (name)
    len = snprintf(out, sizeof(out), "TIME %s\n", name);
  else
    len = snprintf(out, sizeof(out), "TIME\n");
  ecore_con_server_send(server, out, len);
}

void  
echoes_irc_info(Ecore_Con_Server *server, const char *name)
{
  int len=0;
  char out[BUF_MAX];

  if (!server) return;
  if (name)
    len = snprintf(out, sizeof(out), "INFO %s\n", name);
  else
    len = snprintf(out, sizeof(out), "INFO\n");
  ecore_con_server_send(server, out, len);
}

void  
echoes_irc_who(Ecore_Con_Server *server, const char *name, const char *o)
{
  int len=0;
  char out[BUF_MAX];

  if (!server) return;
  if (name && o)
    len = snprintf(out, sizeof(out), "WHO %s %s\n", name, o);
  else if (name)
    len = snprintf(out, sizeof(out), "WHO %s\n", name);
  else 
    len = snprintf(out, sizeof(out), "WHO\n");
  ecore_con_server_send(server, out, len);
}

void  
echoes_irc_whois(Ecore_Con_Server *server, const char *nickmask)
{
  int len=0;
  char out[BUF_MAX];

  if (!server || !nickmask) return;
  len = snprintf(out, sizeof(out), "WHOIS %s\n", nickmask);
  ecore_con_server_send(server, out, len);
}

void  
echoes_irc_whowas(Ecore_Con_Server *server, const char *nick, 
                const char *count, const char *name)
{
  int len=0;
  char out[BUF_MAX];

  if (!server || !nick) return;
  if (nick && count && name)
    len = snprintf(out, sizeof(out), "WHOWAS %s %s %s\n", nick, count, name);
  else if (nick && count)
    len = snprintf(out, sizeof(out), "WHOWAS %s %s\n", nick, count);
  else 
    len = snprintf(out, sizeof(out), "WHOWAS %s\n", nick);
  ecore_con_server_send(server, out, len);
}
