#include "echoes.h"
#include "echoes_ui.h"

static char *_etk_textblock_strescape(const char *str);
static void  _echoes_ui_nick_list_append_cb(void *value, void *data);

void
echoes_ui_nick_set(Echoes_Server *es, const char *nick)
{
  if (!es) return;
  etk_button_label_set(ETK_BUTTON(es->nick_button), nick);
}

const char *
echoes_ui_user_msg_get(Echoes *echoes)
{
  Echoes_Server *es;

  es = echoes_ui_server_get(echoes);
  if (!es) 
    return etk_entry_text_get(ETK_ENTRY(echoes->msg_entry));
  else
    return etk_entry_text_get(ETK_ENTRY(es->msg_entry));
}

void
echoes_ui_user_msg_clear(Echoes *echoes)
{
  Echoes_Server *es;

  es = echoes_ui_server_get(echoes);
  if (!es) 
    etk_entry_clear(ETK_ENTRY(echoes->msg_entry));
  else
    etk_entry_clear(ETK_ENTRY(es->msg_entry));
}

void
echoes_ui_user_msg_set(Echoes *echoes, const char *msg)
{
  Echoes_Server *es;

  es = echoes_ui_server_get(echoes);
  if (!es) 
    etk_entry_text_set(ETK_ENTRY(echoes->msg_entry), msg);
  else
    etk_entry_text_set(ETK_ENTRY(es->msg_entry), msg);
}

void
echoes_ui_server_show(Echoes_Server *es)
{
  etk_toggle_button_active_set(ETK_TOGGLE_BUTTON(es->page->tab), ETK_TRUE);
}

void
echoes_ui_server_add(Echoes *echoes, const char *name, Echoes_Server *es)
{
  Etk_Tree *tree;
  Etk_Tree_Col *col;
  Etk_Tree_Row *row, *parent = NULL;
  int page;

  tree = ETK_TREE(echoes->server_list);
  col = etk_tree_nth_col_get(tree, 0);

  etk_tree_freeze(tree);
  row = etk_tree_selected_row_get(tree);
  parent = etk_tree_row_parent_get(row);
  if (!parent) parent = row;
  row = etk_tree_row_append(tree, parent, col, NULL, NULL, name, NULL);
  etk_tree_row_unfold(parent);
  etk_tree_thaw(tree);

  Etk_Widget *scrolled_view;
  Etk_Widget *textview;
  Etk_Widget *msg_entry;
  Etk_Widget *nick_button;
  Etk_Widget *vbox;
  Etk_Widget *hbox;

  scrolled_view = etk_scrolled_view_new();
  textview = etk_text_view_new();
  msg_entry = etk_entry_new();
  nick_button = etk_button_new_with_label(es->nick);
  vbox = etk_vbox_new(ETK_FALSE, 10);
  hbox = etk_hbox_new(ETK_FALSE, 3);


  Etk_Widget *frame;
  frame = etk_frame_new(name);
  etk_container_add(ETK_CONTAINER(frame), scrolled_view);
  etk_container_add(ETK_CONTAINER(scrolled_view), textview);
  etk_box_append(ETK_BOX(vbox), frame, ETK_BOX_START, ETK_BOX_EXPAND_FILL, 0);
  etk_box_append(ETK_BOX(vbox), hbox, ETK_BOX_START, ETK_BOX_NONE, 0);
  etk_box_append(ETK_BOX(hbox), nick_button, ETK_BOX_START, ETK_BOX_NONE, 0);
  etk_box_append(ETK_BOX(hbox), msg_entry, ETK_BOX_START, ETK_BOX_EXPAND_FILL, 0);
  page = etk_notebook_page_append(ETK_NOTEBOOK(echoes->notebook), name, vbox);
  etk_widget_show_all(vbox);
  etk_widget_focus(msg_entry);
  etk_textblock_object_cursor_visible_set(ETK_TEXT_VIEW(textview)->textblock_object, ETK_FALSE);
  etk_widget_focusable_set(textview, ETK_FALSE);

  es->scrolled_view = scrolled_view;
  es->textview = textview;
  es->row = row;
  es->page = evas_list_nth(ETK_NOTEBOOK(echoes->notebook)->pages, page);
  es->msg_entry = msg_entry;
  es->nick_button = nick_button;
  es->frame = frame;
  etk_tree_row_data_set(row, es);

  etk_toggle_button_active_set(ETK_TOGGLE_BUTTON(es->page->tab), ETK_TRUE);
  etk_tree_row_select(row);

  etk_signal_connect("key-down", ETK_OBJECT(msg_entry), 
                     ETK_CALLBACK(on_msg_entry_key_down_event), echoes);
}

void 
echoes_ui_server_del(Echoes *echoes, Echoes_Server *es)
{
  Etk_Tree *tree;
  Etk_Tree_Row *row_child;
  Echoes_Server *es_child;
  Etk_Widget *page;
  Evas_List *l;
  int i;

  tree = ETK_TREE(echoes->server_list);
  etk_tree_freeze(tree);

  row_child = etk_tree_row_first_child_get(es->row);
  while (row_child)
    {
      es_child = etk_tree_row_data_get(row_child);
      for (l = ETK_NOTEBOOK(echoes->notebook)->pages, i = 0; l; l = l->next, i++)
        {
          if (l->data == es_child->page) break;
        }
      page = etk_notebook_page_remove(ETK_NOTEBOOK(echoes->notebook), i);
      etk_object_destroy(ETK_OBJECT(page));
      row_child = etk_tree_row_next_get(row_child);
    }

  if (!etk_tree_row_walk_prev(es->row, ETK_TRUE))
    etk_notebook_current_page_set(ETK_NOTEBOOK(echoes->notebook), 0);
  else
    etk_tree_row_select(etk_tree_row_walk_prev(es->row, ETK_TRUE));
  etk_tree_row_delete(es->row);
  etk_tree_thaw(tree);

  /*
  for (l = ETK_NOTEBOOK(echoes->notebook)->pages, i = 0; l; l = l->next, i++)
    {
      if (l->data == es->page) 
        {
          page = etk_notebook_page_remove(ETK_NOTEBOOK(echoes->notebook), i);
          etk_object_destroy(ETK_OBJECT(page));
        }
    }
    */
}

Echoes_Server *
echoes_ui_server_get(Echoes *echoes)
{
  Etk_Tree *tree;
  Etk_Tree_Row *row;

  tree = ETK_TREE(echoes->server_list);
  row = etk_tree_selected_row_get(tree);
  if (!row) return NULL;

  return etk_tree_row_data_get(row);
}

void
echoes_ui_serv_msg(Echoes_Server *es, char *msg)
{
  char *txt;
  char *parsed_msg;

  parsed_msg = _etk_textblock_strescape(msg);
  asprintf(&txt, "\n" IRC_SERV_MSG, parsed_msg);
  echoes_ui_direct_write_txt(es, txt);
  free(txt);
  free(parsed_msg);
}

void
echoes_ui_chan_msg(Echoes_Server *es, const char *nick, const char *msg)
{
  char *txt;
  char *parsed_msg;
  char  timestamp[6];
  time_t t;
  int hl = 0;

  t = time(NULL);

  if (!es) return;
  parsed_msg = _etk_textblock_strescape(msg);
  strftime(timestamp, 6, "%H:%M", localtime(&t));
  //TODO un vrai parseur CTCP
  if (!strncmp(msg, "\1ACTION ", 8))
    {
      asprintf(&txt, "\n" IRC_CHAN_ACTION, timestamp, nick, parsed_msg+8);
      txt[strlen(txt)-1] = '\0';
    }
  else if (!strcmp(es->nick, nick))
    asprintf(&txt, "\n" IRC_CHAN_YOU_MSG, timestamp, nick, parsed_msg);
  else
    {
      /* Check if there's an Highlight */
      const char *str = msg;
      while ((str = strcasestr(str, es->nick)))
        {
          /*Skip the nick*/
          str += strlen(es->nick);
          if (*str == ':' || *str == ' ' || *str == ',' 
              || *str == ';' || *str == '.' || *str == '!'
              || *str == '?' || *str == '\0')
            {
              hl = 1;
              break;
            }
        }

      if (hl == 0)
        asprintf(&txt, "\n" IRC_CHAN_SOME_MSG, timestamp, nick, parsed_msg);
      else
        asprintf(&txt, "\n" IRC_CHAN_SOME_HL_MSG, timestamp, nick, parsed_msg);
    }
  echoes_ui_direct_write_txt(es, txt);
  free(txt);
  free(parsed_msg);

  echoes_ui_server_tree_notif(es, ECHOES_STATUS_TALK);
}

void
echoes_ui_join_msg(Echoes_Server *es, char *nick, char *user)
{
  char *txt;

  if (!es) return;
  if (!strcmp(nick, es->nick))
      asprintf(&txt, "\n" IRC_YOU_JOIN, es->name);
  else
      asprintf(&txt, "\n" IRC_SOME_JOIN, nick, user);
  echoes_ui_direct_write_txt(es, txt);
  free(txt);
}

void
echoes_ui_part_msg(Echoes_Server *es, char *nick, char *user, char *msg)
{
  char *txt;

  if (!es) return;
  if (!strcmp(nick, es->nick))
      asprintf(&txt, "\n" IRC_YOU_PART, nick, es->name);
  else
    {
      char *parsed_msg;
      parsed_msg = _etk_textblock_strescape(msg);
      asprintf(&txt, "\n" IRC_SOME_PART, nick, user, parsed_msg);
      free(parsed_msg);
    }
  echoes_ui_direct_write_txt(es, txt);
  free(txt);
}

void 
echoes_ui_nick_msg(Echoes_Server *es, const char *old_nick, const char *new_nick)
{
  char *txt;

  if (!es) return;
  if (!strcasecmp(new_nick, es->nick))
      asprintf(&txt, "\n" IRC_YOU_NICK, new_nick);
  else
      asprintf(&txt, "\n" IRC_SOME_NICK, old_nick, new_nick);
  echoes_ui_direct_write_txt(es, txt);
  free(txt);
}

void
echoes_ui_kick_msg(Echoes_Server *es, char *op, char *nick, char *msg)
{
  char *txt;

  if (!es) return;
  if (!strcmp(nick, es->nick))
    asprintf(&txt, "\n" IRC_YOU_KICK, es->name, op, msg);
  else
    {
      char *parsed_msg;
      parsed_msg = _etk_textblock_strescape(msg);
      asprintf(&txt, "\n" IRC_SOME_KICK, nick, es->name, op, parsed_msg);
      free(parsed_msg);
    }
  echoes_ui_direct_write_txt(es, txt);
  free(txt);
}

void 
echoes_ui_names_msg(Echoes_Server *es)
{
  char *txt, *nick_list;

  if (!es || !es->nicks) return;
  nick_list = echoes_list_to_str(es->nicks);
  asprintf(&txt, "\n" IRC_NAMES, es->name, nick_list);
  echoes_ui_direct_write_txt(es, txt);
  free(txt);
  free(nick_list);
}

void 
echoes_ui_invite_msg(Echoes_Server *es, char *chan, char *op, char *nick)
{
  char *txt;

  if (!es) return;
  if (!strcmp(es->nick, nick))
    {
      asprintf(&txt, "\n" IRC_YOU_ARE_INVITE, chan, op);
      echoes_ui_direct_write_txt(es, txt);
      free(txt);
    }
  else if (!strcmp(es->nick, op))
    {
      asprintf(&txt, "\n" IRC_YOU_MAKE_INVITE, nick, chan);
      echoes_ui_direct_write_txt(es, txt);
      free(txt);
    }
}

void 
echoes_ui_quit_msg(Echoes_Server *es, const char *nick, const char *msg)
{
  char *txt;

  if (!es) return;
  if (!strcmp(es->nick, nick))
    asprintf(&txt, "\n" IRC_YOU_QUIT, msg);
  else
    asprintf(&txt, "\n" IRC_SOME_QUIT, nick, msg);
  echoes_ui_direct_write_txt(es, txt);
  free(txt);
}

void 
echoes_ui_mode_msg(Echoes_Server *es, const char *src, const char *mode, const char *nick)
{
  char *txt;

  if (!es) return;
  if (!nick)
    asprintf(&txt, "\n" IRC_MODE_MSG, src, mode, es->name);
  else
    asprintf(&txt, "\n" IRC_MODE_MSG, src, mode, nick);
  echoes_ui_direct_write_txt(es, txt);
  free(txt);
}

void
echoes_ui_direct_write_txt(Echoes_Server *es, const char *txt)
{
  Etk_Textblock *tb;
  Etk_Textblock_Iter *iter;

  if (!es) return;
  if (!(tb = etk_text_view_textblock_get(ETK_TEXT_VIEW(es->textview))))
    return;

  int w, h;
  int scroll = 0;
  Etk_Range *v;
  
  v = etk_scrolled_view_vscrollbar_get(ETK_SCROLLED_VIEW(es->scrolled_view));
  etk_textblock_object_full_geometry_get(ETK_TEXT_VIEW(es->textview)->textblock_object, 
                                         NULL, NULL, &w, &h);
  if ((ETK_RANGE(v)->value && ETK_RANGE(v)->upper == ETK_RANGE(v)->page_size + ETK_RANGE(v)->value))
    {
      scroll = 1;
    }

  iter = etk_textblock_iter_new(tb);
  etk_textblock_iter_forward_end(iter);
  etk_textblock_insert_markup(tb, iter, txt, -1);

  /* Limit the content of the textblock */
  Etk_String *tb_str;
  tb_str = etk_textblock_text_get(tb, 1);
  etk_textblock_iter_free(iter);

  /* XXX Ranananapété des hack baveux XXX */

  etk_textblock_object_full_geometry_get(ETK_TEXT_VIEW(es->textview)->textblock_object, 
                                         NULL, NULL, &w, &h);
  etk_widget_size_recalc_queue(ETK_WIDGET(es->textview));
  etk_range_range_set(ETK_RANGE(v), ETK_RANGE(v)->lower, h);
  if ((!ETK_WIDGET(v)->visible && (h > ETK_RANGE(v)->page_size)) 
      || scroll)
    {
      etk_range_value_set(ETK_RANGE(v), ETK_RANGE(v)->upper - ETK_RANGE(v)->page_size);
    }
  /* XXX Fin des nanaques baveux XXX */

  echoes_ui_server_tree_notif(es, ECHOES_STATUS_NOTE);
}

void
echoes_ui_server_tree_notif(Echoes_Server *es, Echoes_Server_Status status)
{
  if (!es) return;
  if (es->status == ECHOES_STATUS_TALK && status == ECHOES_STATUS_NOTE) return;
  if (etk_tree_row_is_selected(es->row) && status != ECHOES_STATUS_NONE) return;
  Etk_Tree *tree;
  Etk_Tree_Col *col;
  const char *edj =
   (!access(PACKAGE_SOURCE_DATA_DIR "/pixmaps/machin.edj", R_OK)) ?
   PACKAGE_SOURCE_DATA_DIR "/pixmaps/machin.edj" :
   PACKAGE_DATA_DIR "/pixmaps/machin.edj";
  tree = etk_tree_row_tree_get(es->row);
  col = etk_tree_nth_col_get(tree, 0);
  switch (status)
    {
    case ECHOES_STATUS_DEAD:
      etk_tree_row_fields_set(es->row, ETK_FALSE, col, 
                              edj, "group3", 
                              es->name, NULL);
      break;
    case ECHOES_STATUS_NONE:
      etk_tree_row_fields_set(es->row, ETK_FALSE, col,
                              NULL, NULL,
                              es->name, NULL);
      break;
    case ECHOES_STATUS_NOTE:
      etk_tree_row_fields_set(es->row, ETK_FALSE, col, 
                              edj, "group2", 
                              es->name, NULL);
      break;
    case ECHOES_STATUS_TALK:
      etk_tree_row_fields_set(es->row, ETK_FALSE, col, 
                              edj, "group", 
                              es->name, NULL);
      break;
    }
  es->status = status;
}

void
echoes_ui_nick_fill(Echoes *echoes, Ecore_List *nicks)
{
  Etk_Tree *tree;
  Etk_Tree_Col *col;

  tree = ETK_TREE(echoes->nick_list);
  col = etk_tree_nth_col_get(tree, 0);

  etk_tree_freeze(tree);
  etk_tree_clear(tree);
  if (nicks) ecore_list_for_each(nicks, _echoes_ui_nick_list_append_cb, tree);
  etk_tree_thaw(tree);
}

void
echoes_ui_nick_switch(Echoes *echoes, const char *old_nick, const char *new_nick)
{
  if (echoes_ui_nick_del(echoes, old_nick))
    echoes_ui_nick_add(echoes, new_nick);
}

/* TODO etk_tree_row_insert_sorted */
void
echoes_ui_nick_add(Echoes *echoes, const char *nick)
{
  Etk_Tree *tree;
  Etk_Tree_Col *col;

  tree = ETK_TREE(echoes->nick_list);
  col = etk_tree_nth_col_get(tree, 0);

  etk_tree_freeze(tree);
  etk_tree_row_append(tree, NULL, col, nick, NULL);
  etk_tree_thaw(tree);
}

int
echoes_ui_nick_del(Echoes *echoes, const char *nick)
{
  Etk_Tree *tree;
  Etk_Tree_Col *col;
  Etk_Tree_Row *row;
  char *nick2, *str;

  tree = ETK_TREE(echoes->nick_list);
  col = etk_tree_nth_col_get(tree, 0);

  etk_tree_freeze(tree);
  for (row = etk_tree_first_row_get(tree); row; row = etk_tree_row_next_get(row))
    {
      etk_tree_row_fields_get(row, col, &nick2, NULL);
      str = nick2;
      if (*nick2 == '@' || *nick2 == '&' || *nick2 == '+' ) str++;
      if (!strcmp(str, nick))
        {
          etk_tree_row_delete(row);
          etk_tree_thaw(tree);
          return 1;
        }
    }
  etk_tree_thaw(tree);
  return 0;
}

void
echoes_ui_topic_set(Echoes_Server *es, char *topic)
{
  char *txt;
  char *parsed_topic;

  parsed_topic = _etk_textblock_strescape(topic);
  asprintf(&txt, "\n" IRC_CHAN_TOPIC, es->name, parsed_topic);
  echoes_ui_direct_write_txt(es, txt);
  free(txt);
  free(parsed_topic);
}

void
echoes_ui_topic_end_set(Echoes_Server *es, char *nick, time_t date)
{
  char *txt;

  asprintf(&txt, "\n" IRC_CHAN_TOPIC_END, nick, ctime(&date));
  echoes_ui_direct_write_txt(es, txt);
  free(txt);
}

void
echoes_ui_server_name_update(Echoes_Server *es)
{
  Etk_Tree_Col *col;
  char *image, *key;

  etk_frame_label_set(ETK_FRAME(es->frame), es->name);
  col = etk_tree_nth_col_get(etk_tree_row_tree_get(es->row), 0);
  etk_tree_row_fields_get(es->row, col, &image, &key, NULL, NULL);
  etk_tree_row_fields_set(es->row, ETK_FALSE, col, image, key, es->name, NULL);
}

static char *
_etk_textblock_strescape(const char *str)
{
  char *escaped, *temp;
  char c;
  int  i, j, size;

  if (!str) { return strdup(""); }

  size = strlen(str)+1;
  escaped = malloc(sizeof(char)*size);

  for( i=0, j=0, c=str[0] ; c!='\0' ; i++, j++, c=str[i])
    {
      if(c=='&' || c=='<' || c=='>')
        {
          size += 19;
          temp  = realloc(escaped, size);
          if(!temp)
            {
              free(escaped);
              return strdup("Internal Error");
            }
          escaped = temp;

          strncpy(escaped+j, "<font> &", 8); j+=8;
          switch(c)
            {
            case '&':
              strncpy(escaped+j, "amp", 3); j+=3;
              break;
            case '<':
              strncpy(escaped+j, "lt" , 2); j+=2;
              break;
            case '>':
              strncpy(escaped+j, "gt" , 2); j+=2;
            default:
              break;
            }
          strncpy(escaped+j, ";</font>", 8); j+=7;
        }
      else
        {
          escaped[j] = c;
        }
    }

  escaped[j] = '\0';
  return escaped;
}

static void
_echoes_ui_nick_list_append_cb(void *value, void *data)
{
  char *nick;
  Etk_Tree *tree;

  nick = value;
  tree = data;
  etk_tree_row_append(tree, NULL, etk_tree_nth_col_get(tree, 0), nick, NULL);
}

